// ********************************************************************************
//		This is an iptables rules timeout daemon. It reads commands from a telnet
// or the like connection.
//
// The commands are:
//		add ip-address [[duration] [ port]]
//		del ip-address
//
//	It could be changed to run other commands or be made more generic but that's
// for another time ...
//
// Program options are:
//		-d - debugging
//		-n - the number of 'entries' you want to keep track off
//		-p - the port number it listens on
//		-a - allowed IP's that can talk to it
//
//	To compile it just do: "gcc iptables-timeoutd.c -o iptables-timeoutd"
//
//	This has all the usual caveats, it may or may not work as expected. So, if
// it blows away your server dont blame me ....

/**********************************************************************
 * ###############################################################################
 * #       This program is free software; you can redistribute it and/or
 * #       modify it under the terms of the GNU General Public License
 * #       as published by the Free Software Foundation; either version 2
 * #       of the License, or (at your option) any later version.
 * #
 * #       This program is distributed in the hope that it will be useful,
 * #       but WITHOUT ANY WARRANTY; without even the implied warranty of
 * #       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * #       GNU General Public License for more details.
 * #
 * #       You should have received a copy of the GNU General Public License
 * #       along with this program; if not, write to the Free Software
 * #       Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * ###############################################################################
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <time.h>

// ********************************************************************************
typedef struct iptables_timeout {
	unsigned char      it_ip_address[4];	// the ip address to ban
	unsigned short int it_port;				// the port to ban ... 0 = all
	unsigned short int it_duration;			// the duration left in seconds
} iptables_timeout;

#define PORT		      5667
#define NENTRIES	      5000
#define DEFAULT_TIMEOUT	120
#define NCOMMANDS			40
#define NCONNECTS			10
#define BUFSIZE			1024

extern char *optarg;
extern int optind, opterr, optopt;

iptables_timeout blocked_list[NENTRIES];
int listen_port = PORT, nentries = NENTRIES, default_timeout = DEFAULT_TIMEOUT, debug = 0, connected_from[NCONNECTS];
int port_socket;
char mybuf[BUFSIZE], allowed_ips[NCONNECTS][4];
char *commands[NCOMMANDS+1][100];
struct tm *my_tp;
time_t my_t;

// ********************************************************************************
void myperror(char *msg) {
	fprintf(stdout, "Error: %s %d = %s\n", msg, errno, strerror(errno));
}

// ********************************************************************************
int add_allowed(char *ipa) {
 char ipaddr[4] = {0,0,0,0}, *cp;
 int i;

		// break the IP address apart
	cp = strtok(ipa, "."); if (cp != NULL) { ipaddr[0] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[1] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[2] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[3] = atoi(cp); }

	for (i=0; i<NCONNECTS; i++) {
		if (allowed_ips[i][0] == 0) {
			allowed_ips[i][0] = ipaddr[0];
			allowed_ips[i][1] = ipaddr[1];
			allowed_ips[i][2] = ipaddr[2];
			allowed_ips[i][3] = ipaddr[3];
			break;
		}
	}
	return 1;
}

// ********************************************************************************
int check_allowed(char *ipa) {
 char ipaddr[4] = {0,0,0,0}, *cp;
 int i;

		// break the IP address apart
	cp = strtok(ipa, "."); if (cp != NULL) { ipaddr[0] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[1] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[2] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[3] = atoi(cp); }

	for (i=0; i<NCONNECTS; i++) {
		if (allowed_ips[i][0] == ipaddr[0]) {
			if (allowed_ips[i][1] == ipaddr[1]) {
				if (allowed_ips[i][2] == ipaddr[2]) {
					if (allowed_ips[i][3] == ipaddr[3]) {
						return 1;
					}
				}
			}
		}
	}
	return -1;
}

// ********************************************************************************
int open_port(int port) {
 int sockfd, i=1;
 struct hostent *host;
 struct sockaddr_in soc_addr;

	if ((sockfd = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
if (debug) perror("at socket");
		return -1;
	}

	memset(soc_addr.sin_zero, 0, 8);
	soc_addr.sin_family = PF_INET;
	soc_addr.sin_port = htons(port);
	soc_addr.sin_addr.s_addr = htonl(INADDR_ANY);

	if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (void *)&i, sizeof(int)) == -1) {
if (debug) myperror("at setsockopt");
		return -2;
	}

	if (bind(sockfd, (struct sockaddr *)&soc_addr, sizeof(struct sockaddr)) == -1) {
if (debug) myperror("at bind");
		return -3;
	}

	if (listen(sockfd, 10) == -1) {
if (debug) myperror("at listen");
		return -4;
	}

		// make sure the listening is non-blocking so that we can keep going if there's
		// nothing to read
	fcntl(sockfd, F_SETFL, O_NONBLOCK);

	return sockfd;
}

// ********************************************************************************
void close_port(int sockfd) {
	close(sockfd);
}

// ********************************************************************************
void write_port(int sockfd, char *string) {
 write(sockfd, string, strlen(string));
}

// ********************************************************************************
int read_port(int sockfd)
{
 struct sockaddr_in remote_addr;
 int fd_size = sizeof(struct sockaddr_in);
 int newfd, i, j, k, l;
 char sbuf[BUFSIZE];

	if ((newfd = accept(sockfd, (struct sockaddr *)&remote_addr, &fd_size)) == -1) {
		if (errno != EAGAIN) {
if (debug) myperror("at accept");
			return -1;
		}
	} else {
			// find an empty connection slot to use if it's allowed to connect
		if (check_allowed(inet_ntoa(remote_addr.sin_addr)) != -1) {
				// make sure the listening is non-blocking so that we can keep going if there's
				// nothing to read
			fcntl(newfd, F_SETFL, O_NONBLOCK);

			for (i=0; i<NCONNECTS; i++) {
				if (connected_from[i] == -1) {
					connected_from[i] = newfd;
if (debug) printf("Accepted connection: %d - %s\n", i, inet_ntoa(remote_addr.sin_addr));
					break;
				}
			}
				// find a dead connection and assign it to the new one
			if (i >= NCONNECTS) {
				for (i=0; i<NCONNECTS; i++) {
					if (connected_from[i] != -1) {
						if (send(connected_from[i], "\n", 1, 0) == -1) {
							connected_from[i] = newfd;
if (debug) printf("Accepted connection: %d - %s\n", i, inet_ntoa(remote_addr.sin_addr));
							break;
						}
					}
				}
				if (i >= NCONNECTS) {
if (debug) printf("Too many hosts connected to the daemon !!!\n");
					write_port(newfd, "Too many hosts connected to the daemon ... Bye !!!\n");
					close_port(newfd);
				}
			}
		} else {
			close_port(newfd);
if (debug) printf("Illegal connection: %d - %s\n", i, inet_ntoa(remote_addr.sin_addr));
		}
	}

		// read from each connection slot and get the commands
	l = 0;
	for (i=0; i<NCONNECTS; i++) {
		if (connected_from[i] != -1) {
			if ((j = read(connected_from[i], sbuf, BUFSIZE)) > 0) {
				sbuf[j] = '\0';
				k = 0;
				for (j=0; j<BUFSIZE; j++) {
					if (sbuf[j] == '\0') { 
						if (l < (NCOMMANDS - 1)) {
							strncpy(&commands[l][0], &sbuf[k], 99);
							commands[l][99] = '\0';
							commands[++l][0] = '\0';
						}
						break;
					}
						// break each CR delimited command into separate commands
					if (sbuf[j] == '\n') {
						sbuf[j] = '\0';
						if (l < (NCOMMANDS - 1)) {
							strncpy(&commands[l][0], &sbuf[k], 99);
							commands[l][99] = '\0';
							commands[++l][0] = '\0';
						}
						k = j+1;
					}
				}
//if (debug) printf("Read: %d:%d = %s\n", i, connected_from[i], sbuf);
			}
		}
	}
	return l;
}

// ********************************************************************************
int add_rule(char *ip, int port, int dur) {
 char ipaddr[4] = {0,0,0,0}, *cp;
 int i;

		// break the IP address apart
	cp = strtok(ip, "."); if (cp != NULL) { ipaddr[0] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[1] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[2] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[3] = atoi(cp); }

		// find an empty slot to put it
	for (i=0; i<NENTRIES; i++) {
		if (blocked_list[i].it_ip_address[0] == 0) {
if (debug) printf("Add: %d.%d.%d.%d %d %d @ %d:%d:%d\n", ipaddr[0], ipaddr[1], ipaddr[2], ipaddr[3], port, dur, my_tp->tm_hour, my_tp->tm_min, my_tp->tm_sec);
			blocked_list[i].it_ip_address[0] = ipaddr[0];
			blocked_list[i].it_ip_address[1] = ipaddr[1];
			blocked_list[i].it_ip_address[2] = ipaddr[2];
			blocked_list[i].it_ip_address[3] = ipaddr[3];
			blocked_list[i].it_port = port;
			blocked_list[i].it_duration = dur;

				//*****************************************************************
				// run the command
			if (port == 0) {
				sprintf(mybuf, "iptables -A RH-Firewall-1-INPUT -m state --state NEW -m tcp -p tcp -s %d.%d.%d.%d -j REJECT", ipaddr[0], ipaddr[1], ipaddr[2], ipaddr[3]);
			} else {
				sprintf(mybuf, "iptables -A RH-Firewall-1-INPUT -m state --state NEW -m tcp -p tcp -s %d.%d.%d.%d --dport %d -j REJECT", ipaddr[0], ipaddr[1], ipaddr[2], ipaddr[3], port);
			}
if (debug) printf("Cmd = %s\n", mybuf);
			system(mybuf);
			return 1;
		}
	}

	return -1;
}

// ********************************************************************************
int remove_rule(char *ip) {
 char ipaddr[4] = {0,0,0,0}, *cp;
 int i;

		// break the IP address apart
	cp = strtok(ip, "."); if (cp != NULL) { ipaddr[0] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[1] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[2] = atoi(cp); }
	cp = strtok(NULL, "."); if (cp != NULL) { ipaddr[3] = atoi(cp); }

		// find the slot and clear it
	for (i=0; i<NENTRIES; i++) {
		if (blocked_list[i].it_ip_address[0] == ipaddr[0]) {
			if (blocked_list[i].it_ip_address[1] == ipaddr[1]) {
				if (blocked_list[i].it_ip_address[2] == ipaddr[2]) {
					if (blocked_list[i].it_ip_address[3] == ipaddr[3]) {
if (debug) printf("Del: %s @ %d:%d:%d\n", ip, my_tp->tm_hour, my_tp->tm_min, my_tp->tm_sec);

							//*****************************************************************
							// run the reset command
						if (blocked_list[i].it_port == 0) {
							sprintf(mybuf, "iptables -D RH-Firewall-1-INPUT -m state --state NEW -m tcp -p tcp -s %d.%d.%d.%d -j REJECT", blocked_list[i].it_ip_address[0], blocked_list[i].it_ip_address[1], blocked_list[i].it_ip_address[2], blocked_list[i].it_ip_address[3]);
						} else {
							sprintf(mybuf, "iptables -D RH-Firewall-1-INPUT -m state --state NEW -m tcp -p tcp -s %d.%d.%d.%d --dport %d -j REJECT", blocked_list[i].it_ip_address[0], blocked_list[i].it_ip_address[1], blocked_list[i].it_ip_address[2], blocked_list[i].it_ip_address[3], blocked_list[i].it_port);
						}
if (debug) printf("Cmd = %s\n", mybuf);
						system(mybuf);
						blocked_list[i].it_ip_address[0] = 0;
						return 1;
					}
				}
			}
		}
	}

	return -1;
}

// ********************************************************************************
void timeout_rule() {
 int i;

		// find any ready to be timed out
	for (i=0; i<NENTRIES; i++) {
		if (blocked_list[i].it_ip_address[0] != 0) {
			blocked_list[i].it_duration--;
			if (blocked_list[i].it_duration == 0) {
if (debug) printf("Timeout: %d.%d.%d.%d %d %d @ %d:%d:%d\n", blocked_list[i].it_ip_address[0], blocked_list[i].it_ip_address[1], blocked_list[i].it_ip_address[2], blocked_list[i].it_ip_address[3], blocked_list[i].it_port, blocked_list[i].it_duration, my_tp->tm_hour, my_tp->tm_min, my_tp->tm_sec);

					//*****************************************************************
					// run the reset command
				if (blocked_list[i].it_port == 0) {
					sprintf(mybuf, "iptables -D RH-Firewall-1-INPUT -m state --state NEW -m tcp -p tcp -s %d.%d.%d.%d -j REJECT", blocked_list[i].it_ip_address[0], blocked_list[i].it_ip_address[1], blocked_list[i].it_ip_address[2], blocked_list[i].it_ip_address[3]);
				} else {
					sprintf(mybuf, "iptables -D RH-Firewall-1-INPUT -m state --state NEW -m tcp -p tcp -s %d.%d.%d.%d --dport %d -j REJECT", blocked_list[i].it_ip_address[0], blocked_list[i].it_ip_address[1], blocked_list[i].it_ip_address[2], blocked_list[i].it_ip_address[3], blocked_list[i].it_port);
				}
if (debug) printf("Cmd = %s\n", mybuf);
				system(mybuf);
				blocked_list[i].it_ip_address[0] = 0;
			}
		}
	}
}

// ********************************************************************************
int main(int argc, char *argv[]) { 
 char c, *cp, *ip, ipaddr[20];
 int i, j, dur, port;

		// now, should it do a /etc/init.d/iptables start to reset the rules so that any
		// have been left over from a crash or the like are cleared ???
	//system("/etc/init.d/iptables start");

		// set initial variable values
	memset(&blocked_list[0].it_ip_address[0], '\0', sizeof(iptables_timeout) * nentries);
	for (i=0; i<NCONNECTS; i++) {
		connected_from[i] = -1;
		allowed_ips[i][0] = 0;
	}

		// process any program arg's
	while ((c = getopt(argc, argv, "a:dn:p:")) != -1) {
if (debug) printf("O = %c:%s\n", c, optarg);
		switch (c) {
			case 'a': add_allowed(optarg);
						break;
			case 'd': debug = 1;
						break;
			case 'n': nentries = atoi(optarg);
						break;
			case 'p': listen_port = atoi(optarg);
						break;
		}
	}

	if ((port_socket = open_port(listen_port)) < 0) {
		printf("Unable to open port %d . Error %d\n", port, port_socket);
		exit(1);
	}

		// the main loop
	while (1) {
		if ((i = read_port(port_socket)) > 0) {
			for (j=0; j<(i-1); j++) {
//				my_tp->tm_hour = my_tp->tm_min = my_tp->tm_sec = 0;
				if (time(&my_t) != -1) { my_tp = localtime(&my_t); }

				cp = strtok(&commands[j][0], " ");
				if ((ip = strtok(NULL, " ")) != NULL) { strncpy(ipaddr, ip, 19); ipaddr[19] = '\0'; } else { ipaddr[0] = '\0'; }
if (debug) printf("Cmd = %s\n", ipaddr);

				if (strncasecmp(cp, "add", 3) == 0) {
					port = 0; dur = DEFAULT_TIMEOUT;
					if (ipaddr[0] != '\0') {
						if ((cp = strtok(NULL, " ")) != NULL) { dur = atoi(cp); }
						if ((cp = strtok(NULL, " ")) != NULL) { port = atoi(cp); }
						add_rule(ip, port, dur);
					}
				} else if (strncasecmp(cp, "del", 3) == 0) {
					if (ipaddr[0] != '\0') {
						remove_rule(ip);
					}
				}
			}
//		} else {
//if (debug) printf("No commands read .... \n");
		}
		sleep(1);
//		my_tp->tm_hour = my_tp->tm_min = my_tp->tm_sec = 0;
		if (time(&my_t) != -1) { my_tp = localtime(&my_t); }
		timeout_rule();
	}

	close_port(port_socket);
}
